"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var utils_1 = require("./utils");
var matchesState_1 = require("./matchesState");
var mapState_1 = require("./mapState");
var State_1 = require("./State");
var history_1 = require("./history");
var STATE_DELIMITER = '.';
var HISTORY_KEY = '$history';
function getNextStateValue(parent, stateValue, action, history) {
    if (history === void 0) { history = parent.history; }
    if (typeof stateValue === 'string') {
        var state = parent.states[stateValue];
        var initialState = state.getInitialState();
        if (initialState) {
            stateValue = (_a = {},
                _a[stateValue] = initialState,
                _a);
        }
        else {
            return state.next(action, history) || undefined;
        }
    }
    if (parent.parallel) {
        var initialState = parent.getInitialState();
        if (typeof initialState !== 'string') {
            stateValue = __assign({}, initialState, stateValue);
        }
    }
    if (Object.keys(stateValue).length === 1) {
        var subStateKey = Object.keys(stateValue)[0];
        var subState = parent.states[subStateKey];
        var subStateValue = stateValue[subStateKey];
        var subHistory = history[subStateKey];
        var nextValue_1 = getNextStateValue(subState, subStateValue, action, subHistory);
        if (nextValue_1) {
            return _b = {}, _b[subStateKey] = nextValue_1, _b;
        }
        return subState.next(action, history);
    }
    var nextValue = {};
    var willTransition = false;
    var untransitionedKeys = {};
    Object.keys(stateValue).forEach(function (key) {
        var subValue = getNextStateValue(parent.states[key], stateValue[key], action, history[key]);
        if (subValue) {
            nextValue[key] = subValue;
            willTransition = true;
        }
        else {
            nextValue[key] = undefined;
            untransitionedKeys[key] = stateValue[key];
        }
    });
    return willTransition
        ? Object.assign(nextValue, untransitionedKeys)
        : undefined;
    var _a, _b;
}
var StateNode = /** @class */ (function () {
    function StateNode(config, history) {
        var _this = this;
        this._relativeValue = new Map();
        this.key = config.key;
        this.parent = config.parent;
        this.id = this.parent
            ? this.parent.id + STATE_DELIMITER + this.key
            : this.key;
        this.initial = config.initial;
        this.parallel = !!config.parallel;
        this.history = history || history_1.createHistory(config);
        this.states = config.states
            ? utils_1.mapValues(config.states, function (stateConfig, key) {
                return new StateNode(__assign({}, stateConfig, { key: key, parent: _this }), history);
            })
            : {};
        this.on = config.on;
    }
    StateNode.prototype.transition = function (state, action) {
        var stateValue = (state instanceof State_1.default ? state.value : state) || this.getInitialState();
        var history = state instanceof State_1.default ? state.history : this.history;
        stateValue = utils_1.toTrie(stateValue);
        var nextValue = getNextStateValue(this, stateValue, action, history) ||
            getNextStateValue(this, stateValue, undefined, history);
        return new State_1.default({
            value: nextValue,
            history: history_1.updateHistory(history, stateValue),
            changed: true
        });
    };
    StateNode.prototype.next = function (action, history) {
        if (!action) {
            return this.key;
        }
        var actionType = utils_1.getActionType(action);
        if (!this.on || !this.on[actionType]) {
            return undefined;
        }
        var nextPath = utils_1.toStatePath(this.on[actionType]);
        var currentState = this.parent;
        var currentHistory = history;
        nextPath.forEach(function (subPath) {
            if (subPath === HISTORY_KEY) {
                subPath = currentHistory.$current;
            }
            if (typeof subPath === 'object') {
                subPath = Object.keys(subPath)[0];
            }
            currentState = currentState.states[subPath];
            currentHistory = currentHistory[subPath];
        });
        while (currentState.initial) {
            currentState = currentState.states[currentState.initial];
        }
        return currentState.getRelativeValue(this.parent);
    };
    StateNode.prototype.getInitialState = function () {
        var initialState = this._initialState;
        if (initialState) {
            return initialState;
        }
        initialState = this.parallel
            ? utils_1.mapValues(this.states, function (state) { return state.getInitialState(); })
            : this.initial;
        return (this._initialState = initialState);
    };
    StateNode.prototype.getState = function (relativeStateId) {
        var statePath = utils_1.toStatePath(relativeStateId);
        try {
            return statePath.reduce(function (subState, subPath) {
                return subState.states[subPath];
            }, this);
        }
        catch (e) {
            return undefined;
        }
    };
    Object.defineProperty(StateNode.prototype, "events", {
        get: function () {
            var _this = this;
            if (this._events) {
                return this._events;
            }
            var events = new Set(this.on ? Object.keys(this.on) : undefined);
            Object.keys(this.states).forEach(function (stateId) {
                var state = _this.states[stateId];
                if (state.states) {
                    for (var _i = 0, _a = state.events; _i < _a.length; _i++) {
                        var event = _a[_i];
                        events.add(event);
                    }
                }
            });
            return (this._events = Array.from(events));
        },
        enumerable: true,
        configurable: true
    });
    StateNode.prototype.getRelativeValue = function (toNode) {
        var memoizedRelativeValue = this._relativeValue.get(toNode);
        if (memoizedRelativeValue) {
            return memoizedRelativeValue;
        }
        var initialState = this.getInitialState();
        var relativeValue = initialState
            ? (_a = {},
                _a[this.key] = initialState,
                _a) : this.key;
        var currentNode = this.parent;
        while (currentNode && currentNode !== toNode) {
            var currentInitialState = currentNode.getInitialState();
            relativeValue = (_b = {},
                _b[currentNode.key] = typeof currentInitialState === 'object' &&
                    typeof relativeValue === 'object'
                    ? __assign({}, currentInitialState, relativeValue) : relativeValue,
                _b);
            currentNode = currentNode.parent;
        }
        this._relativeValue.set(toNode, relativeValue);
        return relativeValue;
        var _a, _b;
    };
    return StateNode;
}());
exports.StateNode = StateNode;
function Machine(config) {
    return new StateNode(config);
}
exports.default = { StateNode: StateNode, Machine: Machine, State: State_1.default, matchesState: matchesState_1.default, mapState: mapState_1.default };
