"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var State_1 = require("./State");
function assoc(coll, key, value) {
    if (coll[key] === value) {
        return coll;
    }
    var newColl = Object.assign({}, coll);
    newColl[key] = value;
    return newColl;
}
function get(coll, path) {
    return path.reduce(function (acc, subPath) {
        return acc[subPath];
    }, coll);
}
exports.get = get;
function assocIn(coll, path, value) {
    var key0 = path[0];
    if (path.length === 1) {
        // simplest case is a 1-element array.  Just a simple assoc.
        return assoc(coll, key0, value);
    }
    else {
        // break the problem down.  Assoc this object with the first key
        // and the result of assocIn with the rest of the keys
        return assoc(coll, key0, assocIn(coll[key0] || {}, path.slice(1), value));
    }
}
exports.assocIn = assocIn;
function flatMap(collection, iteratee) {
    return collection.reduce(function (acc, item) {
        var result = iteratee(item);
        if (Array.isArray(result)) {
            acc.push.apply(acc, result);
        }
        else {
            acc.push(result);
        }
        return acc;
    }, []);
}
exports.flatMap = flatMap;
function getActionType(action) {
    try {
        return typeof action === 'string' || typeof action === 'number'
            ? "" + action
            : action.type;
    }
    catch (e) {
        throw new Error('Actions must be strings or objects with a string action.type.');
    }
}
exports.getActionType = getActionType;
function toStatePath(stateId) {
    try {
        if (Array.isArray(stateId)) {
            return stateId;
        }
        return stateId.toString().split('.');
    }
    catch (e) {
        throw new Error("'" + stateId + "' is not a valid state path.");
    }
}
exports.toStatePath = toStatePath;
function toTrie(stateValue) {
    if (typeof stateValue === 'object' && !(stateValue instanceof State_1.default)) {
        return stateValue;
    }
    var statePath = toStatePath(stateValue);
    if (statePath.length === 1) {
        return statePath[0];
    }
    var value = {};
    var marker = value;
    for (var i = 0; i < statePath.length - 1; i++) {
        if (i === statePath.length - 2) {
            marker[statePath[i]] = statePath[i + 1];
        }
        else {
            marker[statePath[i]] = {};
            marker = marker[statePath[i]];
        }
    }
    return value;
}
exports.toTrie = toTrie;
function mapValues(collection, iteratee) {
    var result = {};
    Object.keys(collection).forEach(function (key) {
        result[key] = iteratee(collection[key], key, collection);
    });
    return result;
}
exports.mapValues = mapValues;
